"""
擬似イベントを生成するモジュール

```python
from haniwers import mimic

# ジェネレーターとして使う例
for fe in mimic.fake_events_generator(n=100, seed=None, interval="random"):
    # FakeEventオブジェクトを処理する
    print(fe.timestamp)


# 擬似イベントのデータフレームを取得する例
fake_data = mimic.fake_events(100)
```
"""

import random
import time

import pandas as pd
import pendulum

from .daq import RealEvent


class FakeEvent(RealEvent):
    """擬似イベント用のクラス"""

    seed: int | None = None
    """乱数シード。デフォルト値は`None`"""
    is_fake: bool = True
    """データ種類のフラグ。FakeEventオブジェクトは常に``True``に設定"""

    def model_post_init(self, __context) -> None:
        """メンバー変数を初期化する"""
        random.seed(self.seed)
        self.timestamp = pendulum.now()
        self.top = random.randint(0, 10)
        self.mid = random.randint(0, 10)
        self.btm = random.randint(0, 10)
        self.tmp = random.gauss(mu=25.0, sigma=3) + random.gauss()
        self.atm = random.gauss(mu=101000, sigma=1000) + random.gauss()
        self.hmd = random.gauss(mu=50, sigma=15) + random.gauss()

        if self.top > 0:
            self.adc = random.randint(1, 1024)

    def to_mock_string(self) -> str:
        """モックしたデータ列

        時刻が入ってないスペース区切りの文字列

        :Example:

        ```
        top mid btm adc tmp atm hmd
        ```
        """
        values_ = self.to_list_string()[1:]
        mock_string = (" ").join(values_)
        return mock_string


def fake_events_generator(n: int, seed: int | None = None, interval: str | int = "random"):
    """FakeEventを生成

    `FakeEvent`を`n`回生成します。
    `interval`でイベントごとの生成間隔を変更できます。
    実際のデータ取得を模擬するため、`interval="random"`を指定できます。

    :Args:
        - `n (int)`: イベント数
        - `seed (int | None)` : 乱数シード。デフォルトは `None`
        - `interval (int | str)` : 生成する間隔。デフォルトは "random"

    :Yield:
        - `FakeEvent`: 擬似イベント

    :Example:

    ```python
    for fe in fake_events_generator(n=100, seed=None, interval="random"):
        e = fe.model_dump()
        print(e)
    ```

    """
    for _ in range(n):
        yield FakeEvent(seed=seed)
        if isinstance(interval, int):
            time.sleep(interval)
        elif interval in ("random"):
            r = random.randint(1, 5)
            time.sleep(r)


def fake_events(n: int) -> pd.DataFrame:
    """FakeEventのデータフレームを作成

    :Note:
        - `seed=None`、`interval=0`で`fake_events_generator`を実行
        - 結果を`pd.DataFrame`に変換

    :Args:
        - `n (int)`: 生成するFakeEventの数

    :Returns:
        - `data (pd.DataFrame)`: 擬似イベントのデータフレーム

    :Example:

    ```python
    events = fake_events(100)
    events.info()
    ```

    """
    events = []
    for fe in fake_events_generator(n, seed=None, interval=0):
        e = fe.model_dump()
        events.append(e)
    data = pd.DataFrame(events)
    return data



import csv
from pathlib import Path

import numpy as np
import pandas as pd
import pendulum
from loguru import logger
from scipy.optimize import curve_fit
from scipy.special import erfc
from pydantic import BaseModel

from .config import Daq
from .daq import (
    get_savef_with_timestamp,
    set_vth_retry,
    time_daq,
    scan_daq,
)


def scan_threshold(args: Daq, fname: str, duration: int) -> pd.DataFrame:
    events = scan_daq(args, fname, duration)
    # logger.debug(events)
    rows = []
    for event in events:
        rows.append(event.model_dump())
    data = pd.DataFrame(rows)
    # logger.debug(data)
    return data


def scan_threshold_by_channel(daq: Daq, duration: int, ch: int, vth: int) -> list:
    """
    Run threshold scan by channel.

    :Args:
    - `daq (Daq)`: Daqオブジェクトを指定する
    - `duration (int)`: 測定時間（秒）を指定する
    - `ch (int)`: 測定するチャンネル番号を指定する
    - `vth (int)`: スレッショルド値を指定する

    :Returns:
    - `data (list)`: [測定時刻, チャンネル番号, スレッショルド値, イベント数, 気温など]のリスト
    """

    # Try to set the threshold
    if not set_vth_retry(daq, ch, vth, 3):
        msg = f"Failed to set threshold: ch{ch} - {vth}"
        logger.error(msg)
        return []

    # Collect data
    try:
        # fidは7ケタまで使える
        fid = f"{ch:02}_{vth:04}"
        fname = get_savef_with_timestamp(daq, fid)
        rows = scan_threshold(daq, fname, duration)
        counts = len(rows)
        tmp = rows["tmp"].mean()
        atm = rows["atm"].mean()
        hmd = rows["hmd"].mean()
        msg = f"Saved data to: {fname}"
        logger.info(msg)
    except Exception as e:
        msg = f"Failed to collect data due to: {str(e)}"
        logger.error(msg)
        counts = 0
        tmp = 0
        atm = 0
        hmd = 0

    # Save Summary
    now = pendulum.now().to_iso8601_string()
    data = [now, duration, ch, vth, counts, tmp, atm, hmd]
    fname = Path(daq.saved) / daq.fname_scan
    with fname.open("a", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(data)
    msg = f"Added data to: {fname}"
    logger.info(msg)

    return data


def scan_thresholds_in_serial(
    daq: Daq, duration: int, ch: int, thresholds: list[int]
) -> list[list]:
    """
    Run threshold scan for all channels by default.

    :Args:
    - daq (Daq): Daqオブジェクト
    - duration (int): 測定時間（秒）
    - ch (int): チャンネル番号（1 - 3)
    - thresholds (list[int])： 測定するスレッショルド値のリスト

    :Returns:
    - rows (list[list]):  [測定時刻、チャンネル番号、スレッショルド値、イベント数]のリスト
    """

    # Estimated time for scan
    msg = f"Number of points: {len(thresholds)}"
    logger.info(msg)
    estimated_time = len(thresholds) * duration
    msg = f"Estimated time: {estimated_time} sec."
    logger.info(msg)

    # すべてのチャンネルの閾値を高くする
    set_vth_retry(daq, 1, 500, 5)
    set_vth_retry(daq, 2, 500, 5)
    set_vth_retry(daq, 3, 500, 5)

    rows = []
    n = len(thresholds)
    for i, vth in enumerate(thresholds):
        msg = "-" * 40 + f"[{i+1:2d}/{n:2d}: {vth}]"
        logger.info(msg)
        row = scan_threshold_by_channel(daq, duration, ch, vth)
        if row:
            rows.append(row)

    return rows


def scan_thresholds_in_parallel(daq: Daq, duration: int, thresholds: list[int]):
    # Estimated time for scan
    msg = f"Number of points: {len(thresholds)}"
    logger.info(msg)
    estimated_time = len(thresholds) * duration
    msg = f"Estimated time: {estimated_time} sec."
    logger.info(msg)

    n = len(thresholds)
    for i, vth in enumerate(thresholds):
        msg = "-" * 40 + f"[{i+1:2d}/{n:2d}: {vth}]"
        logger.info(msg)
        # すべてのチャンネルの閾値を設定
        set_vth_retry(daq, 1, vth, 5)
        set_vth_retry(daq, 2, vth, 5)
        set_vth_retry(daq, 3, vth, 5)

        try:
            rows = time_daq(daq, duration)
            counts = len(rows)
            tmp = rows["tmp"].mean()
            atm = rows["atm"].mean()
            hmd = rows["hmd"].mean()
            fname = get_savef_with_timestamp(daq, ch)
            rows.to_csv(fname, index=False)
            msg = f"Saved data to: {fname}"
            logger.info(msg)
        except Exception as e:
            pass
        # それっぽいものを作ってる途中


def scan_thresholds(daq: Daq, duration: int, ch: int, thresholds: list[int]) -> list[list]:
    return scan_thresholds_in_serial(daq, duration, ch, thresholds)


def erfc_function(x, a, b, c, d):
    """
    誤差補正関数（Error function complement）。

    スレショルドを計算するためのフィット関数です。

    erfc(x) = 1 - erf(x)

    Parameters
    ----------
    x : input value
    a : height
    b : mean
    c : sigma
    d : intercept
    """
    return a * erfc((x - b) / c) + d


def fit_threshold_by_channel(data: pd.DataFrame, ch: int, func, params: list[float]):
    """誤差補正関数を使ってスレッショルド値を決める

    Args:
        data (pd.DataFrame): スレッショルド測定のデータフレーム
        ch (int): スレッショルドを求めるチャンネル番号
        func (_type_): フィット関数

    Returns:
        pd.DataFrame: 閾値の提案値のデータフレーム
        pd.DataFrame: フィットしたデータフレーム
        pd.DataFrame: フィット曲線のデータフレーム
    """

    # 実行した時刻を取得する
    now = pendulum.now()

    # データフレームのカラム名を確認する
    expected_names = ["time", "duration", "ch", "vth", "events", "tmp", "atm", "hmd"]
    names = list(data.columns)
    assert names == expected_names

    # フィットの準備
    # 1. 該当するチャンネル番号のデータを抽出
    # 2. イベントレートの計算
    # 3. numpy配列に変換
    q = f"ch == {ch}"
    # print(f"----- Query: {q} -----")
    data_q = data.query(q).copy()
    data_q["event_rate"] = data_q["events"] / data_q["duration"]
    x_data = data_q["vth"]
    y_data = data_q["event_rate"]

    # フィットの初期パラメータ
    # TODO: 初期パラメータを外から調整できるようにする
    # params = [10.0, 300.0, 1.0, 1.0]

    # フィット：1回目
    popt, pcov = curve_fit(func, x_data, y_data, p0=params)
    # std = np.sqrt(np.diag(pcov))

    # logger.debug("フィット（1回目）")
    # logger.debug(f"Parameter Optimized  (popt) = {popt}")
    # logger.debug(f"Parameter Covariance (pcov) = {pcov}")
    # logger.debug(f"Parameter Std. Dev.  (std) = {std}")

    # フィット：2回目
    popt, pcov = curve_fit(func, x_data, y_data, p0=popt)
    # std = np.sqrt(np.diag(pcov))
    # logger.debug("フィット（2回目）")
    # logger.debug(f"Parameter Optimized  (popt) = {popt}")
    # logger.debug(f"Parameter Covariance (pcov) = {pcov}")
    # logger.debug(f"Parameter Std. Dev.  (std) = {std}")

    # フィット曲線
    # 1. フィットで得られた値を使って関数（numpy.array）を作成する
    # 2. データフレームに変換して返り値にする
    xmin = x_data.min()
    xmax = x_data.max()

    # logger.debug(xmin)
    # logger.debug(xmax)
    x_fit = np.arange(xmin, xmax, 0.1)
    a, b, c, d = popt
    y_fit = func(x_fit, a, b, c, d)
    data_f = pd.DataFrame(
        {
            "vth": x_fit,
            "event_rate": y_fit,
            "ch": f"fit{ch}",
        }
    )

    # フィット結果を使って閾値を計算する
    # 例：1sigma, 3sigma, 5sigma
    # pd.DataFrameに変換する
    mean, sigma = popt[1], popt[2]
    _thresholds = {
        "timestamp": now,
        "ch": ch,
        "0sigma": [round(mean)],
        "1sigma": [round(mean + 1 * sigma)],
        "3sigma": [round(mean + 3 * sigma)],
        "5sigma": [round(mean + 5 * sigma)],
    }
    thresholds = pd.DataFrame(_thresholds)

    return thresholds, data_q, data_f


def fit_thresholds(data: pd.DataFrame, channels: list[int], params: list[float]) -> pd.DataFrame:
    """スレッショルド

    Args:
        data (pd.DataFrame): スレッショルド測定のデータフレーム
        channels (list[int]): スレッショルドを計算するチャンネル番号

    Returns:
        pd.DataFrame: 計算したスレッショルド値のデータフレーム
    """
    threshold = []
    erfc = erfc_function
    for c in channels:
        _threshold, _, _ = fit_threshold_by_channel(data, ch=c, func=erfc, params=params)
        threshold.append(_threshold)

    thresholds = pd.concat(threshold, ignore_index=True)
    return thresholds



import platform
import sys
import webbrowser
from pathlib import Path

import pandas as pd
import pendulum
import typer
from loguru import logger
from platformdirs import PlatformDirs
from typing_extensions import Annotated

from . import __version__

DOCS = {
    "home": "https://qumasan.gitlab.io/haniwers/docs/",
    "version": "https://qumasan.gitlab.io/haniwers/docs/command/version/",
    "scan": "https://qumasan.gitlab.io/haniwers/docs/command/scan/",
    "fit": "https://qumasan.gitlab.io/haniwers/docs/command/fit/",
    "vth": "https://qumasan.gitlab.io/haniwers/docs/command/vth/",
    "daq": "https://qumasan.gitlab.io/haniwers/docs/command/daq/",
    "raw2tmp": "https://qumasan.gitlab.io/haniwers/docs/command/raw2tmp/",
    "run2csv": "https://qumasan.gitlab.io/haniwers/docs/command/run2csv/",
}
"""URLs of online documents"""


def _setup_logger(level="INFO") -> Path:
    """ロガーの設定"""

    format_short = (" | ").join(
        ["{time:YYYY-MM-DDTHH:mm:ss}", "<level>{level:8}</level>", "<level>{message}</level>"]
    )
    format_long = (" | ").join(
        [
            "{time:YYYY-MM-DDTHH:mm:ss}",
            "<level>{level:8}</level>",
            "<cyan>{name}.{function}:{line}</cyan>",
            "<level>{message}</level>",
        ]
    )

    # ロガーをリセット
    logger.remove()

    # stderr用
    if level in ["DEBUG"]:
        logger.add(
            sys.stderr,
            format=format_long,
            level=level,
        )
    else:
        logger.add(
            sys.stderr,
            format=format_short,
            level=level,
        )

    # ファイル出力用
    p = PlatformDirs(appname="haniwers", version=__version__)
    fname = p.user_log_path / "haniwers_log.json"
    logger.add(
        sink=fname,
        format=format_long,
        level="DEBUG",
        serialize=True,
        retention="10 days",
        rotation="1 MB",
    )
    return fname


def _open_docs(value: bool, url: str) -> None:
    """Open document with browser.

    コマンドの使い方をのドキュメントをブラウザで開く

    :Args:
        - `value (bool)`: boolean
        - `url (str)`: document URL

    :Example:

    既存コマンドに追加した``--docs``オプションのcallback関数として利用

    ```python
    typer.Option(
        "--docs",
        help=f"Open online document. ({DOCS['version']})",
        callback=lambda v: _open_docs(v, DOCS["version"]),
    )
    ```
    """
    if value:
        msg = f"Open docs with browser: {url}"
        logger.info(msg)
        webbrowser.open(url)
        raise typer.Exit()


app = typer.Typer()


@app.command()
def version(
    env: Annotated[bool, typer.Option(help="Show environment.")] = False,
    log_level: Annotated[str, typer.Option(help="Change log level.")] = "INFO",
    docs: Annotated[
        bool,
        typer.Option(
            "--docs",
            help=f"Open online document. ({DOCS['version']})",
            callback=lambda v: _open_docs(v, DOCS["version"]),
        ),
    ] = False,
):
    """Show haniwers version.

    :Args:
    - env (bool, optional): Show with environment details. Defaults to False.
    - log_level (str, optional): Log level. Defaults to "INFO".

    :Example:

    ```console
    $ haniwers version
    haniwers 0.19.1
    ```

    :Example:

    ```
    $ haniwers version --env
    haniwers 0.19.2

    Environments:

    Logs: ~/Library/Logs/haniwers/0.19.2/haniwers_log.json
    Executable: ~/.local/pipx/venvs/haniwers/bin/python
    Python: 3.12.5
    Implementation: CPython
    Compiler: Clang 15.0.0 (clang-1500.3.9.4)
    OS: macOS-14.6.1-arm64-arm-64bit
    System: darwin
    Platform: Darwin
    Kernel: 23.6.0
    Arch: arm64
    ```

    """

    fname = _setup_logger(level=log_level)

    msg = f"haniwers {__version__}"
    print(msg)
    logger.debug(msg)

    if env:
        _envs = {
            "Logs": fname,
            "Executable": sys.executable,
            "Python": platform.python_version(),
            "Implementation": platform.python_implementation(),
            "Compiler": platform.python_compiler(),
            "OS": platform.platform(),
            "System": sys.platform,
            "Platform": platform.system(),
            "Kernel": platform.release(),
            "Arch": platform.machine(),
        }

        print("\nEnvironments:\n")

        for k, v in _envs.items():
            msg = f"  {k}: {v}"
            print(msg)
            logger.debug(msg)

        print("\n")

    return


@app.command()
def docs(
    page: Annotated[
        str,
        typer.Option(
            help="Set page name.",
        ),
    ] = "home",
):
    """Open online document.

    オンラインドキュメントを開きます。

    :VersionAdded: 0.19.0

    :Example:

    ```console
    $ haniwers docs
    $ haniwers docs --page version
    $ haniwers docs --page scan
    $ haniwers docs --page fit
    $ haniwers docs --page vth
    $ haniwers docs --page daq
    $ haniwers docs --page raw2tmp
    $ haniwers docs --page run2csv
    ```

    """
    url = DOCS[page]
    _open_docs(True, url)


@app.command()
def ports(log_level: Annotated[str, typer.Option(help="Change log level")] = "INFO") -> None:
    """Search available ports and show device names.

    接続可能なポート名の一覧を表示します。
    OSECHIが接続されたポート名は、OSごとに以下のように変化します。

    - Linuxの場合: `/dev/ttyUSB0`
    - macOSの場合: `/dev/cu.usbserial-*` （CP2102N USB to UART Bridge Controller）
    - Windowsの場合: `COM3`

    :Args:
        log_level (str, optional): Log level. Defaults to "INFO".

    :Example:

    ```console
    $ haniwers ports
    | INFO     | Found 2 ports
    | INFO     | Port0: /dev/cu.Bluetooth-Incoming-Port - n/a
    | INFO     | Port1: /dev/cu.usbserial-140 - USB Serial
    ```

    """

    _setup_logger(level=log_level)

    from serial.tools import list_ports

    ports = list_ports.comports()
    n = len(ports)

    if n == 0:
        logger.warning("No ports found")
        return

    logger.info(f"Found {n} ports")

    for i, port in enumerate(ports):
        logger.info(f"Port{i}: {port}")

        logger.debug(f"{port.device=}")
        logger.debug(f"{port.name=}")
        logger.debug(f"{port.description=}")
        logger.debug(f"{port.usb_description()=}")
        logger.debug(f"{port.hwid=}")
        logger.debug(f"{port.usb_info()=}")
        logger.debug(f"{port.pid=}")
        logger.debug(f"{port.vid=}")
        logger.debug(f"{port.interface=}")
        logger.debug(f"{port.manufacturer=}")
        logger.debug(f"{port.product=}")
        logger.debug(f"{port.serial_number=}")

    for port in ports:
        if "UART" in port.description:
            logger.info(f"Please use '{port.device}' as your device path")


@app.command()
def scan(
    ch: Annotated[int, typer.Option(help="Set channel ID.")] = 0,
    duration: Annotated[int, typer.Option(help="Set duration. Unit: [sec]")] = 10,
    step: Annotated[int, typer.Option(help="Set step interval. Unit: [step]")] = 1,
    vmin: Annotated[int, typer.Option(help="Set start point. Unit: [step]")] = 250,
    vmax: Annotated[int, typer.Option(help="Set end point. Unit: [step]")] = 311,
    quiet: Annotated[bool, typer.Option(help="Quiet mode.")] = False,
    load_from: Annotated[str, typer.Option(help="Set filename.")] = "scan.toml",
    log_level: Annotated[str, typer.Option(help="Change log level.")] = "INFO",
    docs: Annotated[
        bool,
        typer.Option(
            "--docs",
            help=f"Open online document. ({DOCS['scan']})",
            callback=lambda v: _open_docs(v, DOCS["scan"]),
        ),
    ] = False,
) -> None:
    """スレッショルド測定

    :Args:
        - `ch (int, optional)`: チャンネル番号. Defaults to 0.
        - `duration (int, optional)`: 1点あたりの測定時間（秒）. Defaults to 10.
        - `step (int, optional)`: 測定間隔. Defaults to 1.
        - `vmin (int, optional)`: 測定範囲（最小値）. Defaults to 250.
        - `vmax (int, optional)`: 測定範囲（最大値）. Defaults to 311.
        - `quiet (bool, optional)`: _description_. Defaults to False.
        - `load_from (str, optional)`: 設定ファイル. Defaults to "scan.toml".

    :Example:

    ```console
    $ haniwers scan
    $ haniwers scan --ch 1
    $ haniwers scan --duration=30
    $ haniwers scan --step=5
    $ haniwers scan --step=10 --vmin=200 --vmin=400

    """
    from .config import Daq
    from .threshold import scan_thresholds

    _setup_logger(level=log_level)
    daq = Daq()
    daq.load_toml(load_from)
    daq.quiet = quiet

    now = pendulum.now().format("YYYYMMDD")
    daq.saved = str(Path(daq.saved) / now)

    if ch == 0:
        channels = [1, 2, 3]
    else:
        channels = [ch]

    thresholds = list(range(vmin, vmax, step))

    for ch in channels:
        msg = f"Running threshold scan on ch{ch}."
        logger.info(msg)
        scan_thresholds(daq, duration, ch, thresholds)

    return


@app.command()
def fit(
    read_from: Annotated[str, typer.Argument(help="Set directory.")],
    # p0: Annotated[list, typer.Option(help="Set initial parameters.")] = [10, 300, 1, 1],
    search_pattern: Annotated[str, typer.Option(help="Set filename.")] = "threshold_scan.csv",
    ch: Annotated[int, typer.Option(help="Set channel ID")] = 0,
    log_level: Annotated[str, typer.Option(help="Change log level")] = "INFO",
    docs: Annotated[
        bool,
        typer.Option(
            "--docs",
            help=f"Open online document. ({DOCS['fit']})",
            callback=lambda v: _open_docs(v, DOCS["fit"]),
        ),
    ] = False,
):
    """スレッショルド計算

    スレッショルドを測定したデータから、誤差関数を使って閾値の推定値を計算します。
    チャンネル番号を指定できます。
    デフォルトはすべてのチャンネルの推定値を計算します。
    計算するたびにその結果は``thresholds_history.csv``に追記されます。
    また、最新のフィット結果は``thresholds_latest.csv``に保存されます。
    このファイルを使ってスレッショルドを設定できるようになっています。

    :Args:
        - `read_from (str)`: スレッショルド測定データがあるディレクトリ名を指定してください
        - `search_pattern (str, optional)`: スレッショルド測定データのファイル名を変更できます。Defaults to "threshold_scan.csv".
        - `ch (int, optional)`: チャンネル番号を変更できます。Defaults to 0.

    :Example:

    ```console
    $ haniwers fit ディレクトリ名
    ```
    """
    import pandas as pd

    from .preprocess import get_fnames
    from .threshold import fit_thresholds

    _setup_logger(level=log_level)

    logger.info(f"Read data from {read_from}")
    fnames = get_fnames(read_from, search_pattern)

    # ファイルが見つからない時は、なにもしない
    if len(fnames) == 0:
        logger.error("No files found.")
        return

    logger.debug(fnames)

    # チャンネル番号が範囲外のときは、なにもしない
    if ch > 3:
        logger.error(f"Out of range!: {ch}")
        return

    channels = [ch]
    if ch == 0:
        channels = [1, 2, 3]

    names = ["time", "duration", "ch", "vth", "events", "tmp", "atm", "hmd"]
    data = pd.read_csv(fnames[0], names=names, parse_dates=["time"])
    thresholds = fit_thresholds(data, channels)

    # 実行した時刻を上書きする
    now = pendulum.now()
    thresholds["timestamp"] = now
    print(thresholds)

    fname = "thresholds_history.csv"
    thresholds.to_csv(fname, index=False, mode="a", header=None)
    logger.info(f"Saved to {fname}")
    fname = "thresholds_latest.csv"
    thresholds.to_csv(fname, index=False)
    logger.info(f"Saved to {fname}")

    return


@app.command()
def vth(
    ch: Annotated[int, typer.Option(help="Set channel ID")] = 0,
    vth: Annotated[int, typer.Option(help="Set threshold value.")] = 0,
    max_retry: Annotated[int, typer.Option(help="Set numbers to retry.")] = 3,
    load_from: Annotated[str, typer.Option(help="Set filename.")] = "daq.toml",
    log_level: Annotated[str, typer.Option(help="Change log level.")] = "INFO",
    docs: Annotated[
        bool,
        typer.Option(
            "--docs",
            help=f"Open online document. ({DOCS['vth']})",
            callback=lambda v: _open_docs(v, DOCS["vth"]),
        ),
    ] = False,
) -> None:
    """スレッショルド設定

    各チャンネルにスレッショルドを設定します。
    カレントディレクトリにある ``thresholds_latest.csv`` を参考に、
    スレッショルド値の最良推定値を設定します。
    1step = 4mVに相当します。

    原因がまだわかってないですが、スレッショルドの書き込みに失敗することがあります。
    その場合は、書き込みに成功するまで``max_retry``回リトライします。デフォルトは3回です。

    :Args:
        - `ch (int, optional)`: チャンネル番号. Defaults to 0 (=all).
        - `vth (int, optional)`: スレッショルド値. Defaults to 0 (=all).
        - `max_retry (int, optional)`: スレッショルドの書き込みに失敗したときにリトライする回数. Defaults to 3.
        - `load_from (str, optional)`: 設定ファイル. Defaults to "daq.toml".

    :Example:

    すべてのチャンネルをまとめて設定

    ```console
    $ haniwers vth
    ```

    :Example:

    チャンネルごとに個別に設定

    ```console
    $ haniwers vth --ch 1 --vth 278
    $ haniwers vth --ch 2 --vth 268
    $ haniwers vth --ch 3 --vth 300
    ```

    """

    from .config import Daq
    from .daq import set_vth_retry

    _setup_logger(level=log_level)

    daq = Daq()
    daq.load_toml(load_from)

    now = pendulum.now().format("YYYYMMDD")
    daq.saved = str(Path(daq.saved) / now)

    # 個別のチャンネルにスレッショルドを設定する
    if ch in range(1, 4) and vth > 0:
        logger.debug(f"Set threshold to each channel: {ch} -> {vth}")
        set_vth_retry(daq, ch, vth, max_retry)
        return

    # 引数を指定しない場合は
    # すべてのチャンネルに規定のスレッショルドを設定する
    if ch == 0 and vth == 0:
        # スレッショルド値をファイルから読み込む
        fname = Path("thresholds_latest.csv")
        if not fname.exists():
            logger.error(f"No file found. Please create {fname}")
            return

        names = ["ch", "3sigma"]
        thresholds = pd.read_csv(fname)[names]

        for _, row in thresholds.iterrows():
            ch = int(row["ch"])
            vth = int(row["3sigma"])
            logger.debug(f"Set threshold to channels: {ch} -> {vth}")
            set_vth_retry(daq, ch, vth, max_retry)
        return

    # オプション指定が間違っている
    logger.error("Invalid arguments")
    return


@app.command()
def daq(
    quiet: Annotated[bool, typer.Option(help="Quiet mode.")] = False,
    load_from: Annotated[str, typer.Option(help="Set filename.")] = "daq.toml",
    log_level: Annotated[str, typer.Option(help="Change log level")] = "INFO",
    docs: Annotated[
        bool,
        typer.Option(
            "--docs",
            help=f"Open online document. ({DOCS['daq']})",
            callback=lambda v: _open_docs(v, DOCS["daq"]),
        ),
    ] = False,
) -> None:
    """Start DAQ. Set up with daq.toml.

    :Args:
        - `quiet (bool, optional)`: quiet mode. Defaults to False.
        - `load_from (str, optional)`: 設定ファイル. Defaults to "daq.toml".

    :Example:

    ```console
    $ haniwers daq
    ```
    """
    from .config import Daq
    from .daq import open_serial_connection, run

    # ログレベルを設定
    _setup_logger(level=log_level)

    # DAQの初期設定
    args = Daq()
    args.load_toml(load_from)
    args.quiet = quiet

    # データの保存先をymdに変更
    now = pendulum.now().format("YYYYMMDD")
    args.saved = str(Path(args.saved) / now)

    with open_serial_connection(args) as port:
        run(port, args)

    return


@app.command()
def raw2tmp(
    read_from: Annotated[str, typer.Argument(help="Set directory.")],
    search_pattern: Annotated[str, typer.Option(help="Set search pattern.")] = "*.csv",
    interval: Annotated[int, typer.Option(help="Set resample interval. Unit: [sec]")] = 600,
    offset: Annotated[int, typer.Option(help="Set datetime offset. Unit: [sec]")] = 0,
    tz: Annotated[str, typer.Option(help="Set time-zone.")] = "UTC+09:00",
    log_level: Annotated[str, typer.Option(help="Change log level.")] = "INFO",
    docs: Annotated[
        bool,
        typer.Option(
            "--docs",
            help=f"Open online document. ({DOCS['raw2tmp']})",
            callback=lambda v: _open_docs(v, DOCS["raw2tmp"]),
        ),
    ] = False,
) -> None:
    """Parse raw_data into CSV format. Should be used temporarily for quick analysis.

    宇宙線を測定したデータをその場で確認するために、生データを変換するための簡易コマンドです。
    必要最低限のオプションしか変更できないようになっています。
    出力ファイル名（``tmp_raw2tmp.csv``）は固定で、変更できません。
    きちんとしたデータ解析には ``haniwers run2csv`` を使って変換してください。

    :Args:
        - `read_from (str)`: 測定データがあるディレクトリ名
        - `search_pattern (str, optional)`: 検索パターン. Defaults to "*.csv".
        - `interval (int, optional)`: リサンプル間隔. Defaults to 600.
        - `offset (int, optional)`: 測定時刻のオフセット. Defaults to 0.
        - `tz (str, optional)`: 測定時刻のタイムゾーン. Defaults to "UTC+09:00".

    :Example:

    ```console
    $ haniwers raw2tmp ディレクトリ名
    $ haniwers raw2tmp ディレクトリ名 --search-pattern="*.dat"
    $ haniwers raw2tmp ディレクトリ名 --interval=10
    $ haniwers raw2tmp ディレクトリ名 --offset=36480
    $ haniwers raw2tmp ディレクトリ名 --tz="UTC+09:00"
    ```
    """
    from .preprocess import get_fnames, raw2csv

    _setup_logger(level=log_level)

    logger.info(f"Read data from {read_from}")
    fnames = get_fnames(read_from, search_pattern)
    gzip, csv = raw2csv(fnames, interval, offset, tz)
    logger.debug(f"raw2gz = {len(gzip)}")
    logger.debug(f"raw2csv = {len(csv)}")

    fname = "tmp_raw2tmp.csv.gz"
    gzip.to_csv(fname, index=False, compression="gzip")
    logger.info(f"Save data to: {fname} ({len(gzip)} rows).")

    fname = "tmp_raw2tmp.csv"
    csv.to_csv(fname, index=False)
    logger.info(f"Save data to: {fname} ({len(csv)} rows).")


@app.command()
def run2csv(
    run_id: Annotated[int, typer.Argument(help="Run ID")],
    save: Annotated[bool, typer.Option(help="Set True to save dataframes.")] = False,
    load_from: Annotated[str, typer.Option(help="Set filename.")] = "runs.csv",
    drive: Annotated[str, typer.Option(help="Set directory.")] = "../data",
    log_level: Annotated[str, typer.Option(help="Change log level")] = "INFO",
    docs: Annotated[
        bool,
        typer.Option(
            "--docs",
            help=f"Open online document. ({DOCS['run2csv']})",
            callback=lambda v: _open_docs(v, DOCS["run2csv"]),
        ),
    ] = False,
) -> None:
    """Parse raw_data into CSV format. Specify RunID.

    :Args:
        - `run_id (int)`: ラン番号
        - `save (bool, optional)`: 保存ファイル名（CSV形式）. Defaults to False.
        - `load_from (str, optional)`: RunDataの設定ファイル. Defaults to `"runs.csv"`.
        - `drive (str, optional)`: 設定ファイルがあるディレクトリ. Defaults to `"../data"`.

    :Example:

    ```console
    $ haniwers run2csv 98
    $ haniwers run2csv 98 --save
    $ haniwers run2csv 98 --drive="../data" --load_from="runs.csv"
    ```

    :Example:

    無効なラン番号を指定した場合は、処理せず終了

    ```console
    $ haniwers run2csv 200
    | INFO     | Load config from: runs.csv.
    | INFO     | Get RunData: 200.
    | ERROR    | Run #200 is out of range. Quit.
    | WARNING  | Check or Fix configuration file: ../data/runs.csv.
    ```

    """

    from .config import RunManager
    from .preprocess import run2csv

    _setup_logger(level="INFO")
    rm = RunManager(load_from=load_from, drive=drive)
    msg = f"Load config from: {load_from}."
    logger.info(msg)

    msg = f"Get RunData: {run_id}."
    logger.info(msg)

    run = rm.get_run(run_id)
    logger.info(f"description: {run.description}")
    logger.info(f"read_from: {run.read_from}")
    logger.debug(f"srcf: {run.srcf}")

    gzip, csv = run2csv(run)
    if save:
        fname = run.raw2gz
        gzip.to_csv(fname, index=False, compression="gzip")
        logger.info(f"Save data to: {fname} ({len(gzip)} rows).")
        fname = run.raw2csv
        csv.to_csv(fname, index=False)
        logger.info(f"Save data to: {fname} ({len(csv)} rows).")
    else:
        logger.warning("No data saved. Add --save to save data.")
        logger.debug(f"gzip: {len(gzip)}.")
        logger.debug(f"csv:  {len(csv)}.")


@app.command()
def mock_daq(
    quiet: Annotated[bool, typer.Option(help="Quied mode.")] = False,
    load_from: Annotated[str, typer.Option(help="Set filename")] = "daq.toml",
    log_level: Annotated[str, typer.Option(help="Change log level")] = "DEBUG",
):
    """Dummy DAQ

    シリアル通信をモックして、OSECHIがない場合でもデータを作成できるようにしたコマンド。
    デバッグ用に使う時があるかもしれない。

    """
    from unittest.mock import MagicMock, patch

    import serial

    from .config import Daq
    from .daq import run
    from .mimic import FakeEvent

    # ログレベルを設定
    _setup_logger(level=log_level)
    logger.debug("mock-daq")

    # DAQの初期設定
    args = Daq()
    args.load_toml(load_from)
    args.quiet = quiet

    # データの保存先をymdに変更
    now = pendulum.now().format("YYYYMMDD")
    args.saved = str(Path(args.saved) / now)
    args.max_rows = 10
    args.max_files = 5
    logger.debug(args)

    # シリアル通信をモック
    mock_port = MagicMock()
    mock_port.readline().decode.return_value = FakeEvent().to_mock_string()
    mock_port.name.return_value = "mock"

    with patch("serial.Serial", return_value=mock_port):
        with serial.Serial() as port:
            logger.debug(f"Port opened: {port.name}")
            run(port, args)
    logger.debug(f"Port closed: {port.name}")


if __name__ == "__main__":
    app()



import sys
from pathlib import Path

import pandas as pd
import polars as pl
from loguru import logger


from .config import RunData


def get_fnames(read_from: str, search_pattern: str) -> list[Path]:
    """測定データの一覧を取得

    測定データのファイル名の一覧を取得します。
    ファイル名はPathオブジェクトに変換しています。
    リストはファイル名で昇順ソートします。

    :Args:
        - `read_from (str)`: 測定データがあるパス
        - `search_pattern (str)`: ファイル名の検索パターン

    :Returns:
        - `fnames (list[Path])`: 測定データの一覧

    :Example:

    ```python
    fnames = get_fnames(
        read_from="../data/raw_data/20240601_run80",
        search_pattern="osechi_data_*.csv")
    ```
    """
    fnames = sorted(Path(read_from).glob(search_pattern))
    return fnames


def read_data_with_pandas(fnames: list[Path]) -> pd.DataFrame:
    """pd.DataFrameに変換

    複数の測定ファイルを読み込んでpd.DataFrameに変換します。
    測定ファイルの拡張子は``.dat``と``.csv``に限定しています。

    :Args:
        - `fnames (list[Path])`: 測定データのファイル名のリスト

    :Exception:
        - ``.dat`` / ``.csv``以外の拡張子の場合は中断（break）

    :Returns:
        - `merged (pd.DataFrame)`: 結合したデータフレーム
    """
    names = ["datetime", "top", "mid", "btm", "adc", "tmp", "atm", "hmd"]
    data = []
    for fname in fnames:
        _suffix = fname.suffix
        if _suffix in [".dat"]:
            datum = pd.read_csv(fname, names=names, sep=" ", comment="t")
        elif _suffix in [".csv"]:
            datum = pd.read_csv(fname, names=names)
        else:
            msg = f"Unknown suffix: {_suffix}"
            logger.warning(msg)
            break
        data.append(datum)
    merged = pd.concat(data, ignore_index=True).dropna(how="all")

    if len(merged) == 0:
        msg = "No entries in DataFrame. Exit."
        logger.error(msg)
        sys.exit()

    return merged


def read_data_with_polars(fnames: list[Path]) -> pd.DataFrame:
    """pl.DataFrameに変換

    複数のファイルを読み込んでpl.DataFrameに変換します。
    polarsで読み込んだ方が、少し処理が速くなります。

    ただし、そのままpl.DataFrameを返すと、あとの処理がうまくいかないので、
    最後にpl.DataFrameからpd.DataFrameに変換しています。

    :Args:
        - `fnames (list[Path])`: 測定データのファイル名のリスト

    :Returns:
        - `merged (pd.DataFrame)`: 結合したデータフレーム
    """
    names = ["datetime", "top", "mid", "btm", "adc", "tmp", "atm", "hmd"]
    data = []
    for fname in fnames:
        _suffix = fname.suffix
        _stem = fname.stem
        if _suffix in [".dat"]:
            datum = pl.read_csv(
                fname,
                has_header=False,
                new_columns=names,
                separator=" ",
                comment_char="t",
            )
        elif _suffix in [".csv"]:
            if _stem.startswith("osechi_data"):
                datum = pl.read_csv(fname, has_header=False, new_columns=names)
            elif _stem.startswith("scan_data"):
                datum = pl.read_csv(fname, has_header=False, new_columns=names, skip_rows=1)
        else:
            msg = f"Unknown suffix: {_suffix}"
            logger.warning(msg)
            break
        data.append(datum)
    merged = pl.concat(data).to_pandas().dropna(how="all")

    if len(merged) == 0:
        msg = "No entries in DataFrame. Exit."
        logger.error(msg)
        sys.exit()

    return merged


def read_data(fnames: list[Path]) -> pd.DataFrame:
    """測定データをデータフレームに変換

    ``read_data_with_polars``を使ってデータフレームに変換します。

    :Example:

    ```python
    fnames = get_fnames(
        read_from=...,
        search_pattern=".csv"
        )
    data = read_data(fnames)
    ```
    """
    # data = read_data_with_pandas(fnames)
    data = read_data_with_polars(fnames)
    return data


def add_time(data: pd.DataFrame, offset: int, timezone: str) -> pd.DataFrame:
    """`time`カラムを追加

    測定データのデータフレームに``time``カラムを追加します。
    ``time``カラムは``datetime``カラムをタイムゾーン付き日時オブジェクトに変換したものです。
    また、``offset``（秒）の分、測定時刻を正しい時刻に修正します。

    :Args:
        - `data (pd.DataFrame)`: 測定データ
        - `offset (int)`: 測定時刻のオフセット（秒）
    :Returns:
        - `data (pd.DataFrame)`: ``time``カラムを追加した測定データ

    :Example:

    ```python
    data = read_data(fnames)
    data = add_time(data)
    ```

    ```{caution}
    Raspberry Pi 4はRTCを内蔵していません。電源を切ると時計がストップします。
    測定を開始する前に、ネットワークに接続してNTPから時刻を自動取得するか、
    ``date``コマンドなどで時刻を手動設定する必要があります。
    その作業を忘れてしまった場合のために、``offset``オプションがあります。
    ```

    """
    data["datetime"] = pd.to_datetime(data["datetime"], format="ISO8601")
    if data["datetime"].dt.tz is None:
        data["datetime"] = data["datetime"].dt.tz_localize(timezone)
    data["datetime_offset"] = pd.to_timedelta(offset, "s")
    data["time"] = data["datetime"] + data["datetime_offset"]
    return data


def add_hit(data: pd.DataFrame) -> pd.DataFrame:
    """`hit`カラムを追加

    各シンチのレイヤーにヒットがあったかどうかを確認し、``hit_レイヤー名``カラムを追加します。
    ヒットがある場合は``1``、ヒットがない場合は``0``です。

    :Args:
        - `data (pd.DataFrame)`: 測定データ

    :Returns:
        - `data (pd.DataFrame)`: `hit`カラムを追加したデータフレーム

    :Example:

    ```python
    data = read_data(fnames)
    data = add_time(data)
    data = add_hit(data)
    ```
    """
    copied = data.copy()
    headers = ["top", "mid", "btm"]
    for header in headers:
        name = f"hit_{header}"
        copied[name] = 0
        isT = data[header] > 0
        copied.loc[isT, name] = 1
    return copied


def add_hit_type(data: pd.DataFrame) -> pd.DataFrame:
    """`hit_type`カラムを追加

    レイヤーごとのヒット情報から、ヒットの種類を計算します。
    ヒット情報が計算できてない場合は、警告を表示し、元のデータを返します。

    :Args:
        - `data (pd.DataFrame)`: ``add_hit``したあとの測定データ

    :Returns:
        - `data (pd.DataFrame)`: ``hit_type``を追加したデータフレーム

    :Example:

    ```python
    data = read_data(fnames)
    data = add_time(data)
    data = add_hit(data)
    data = add_hit_type(data)
    ```
    """
    _header = "hit_top"
    if _header not in data.columns:
        msg = f"Header '{_header}' does not exist in current dataframe."
        logger.warning(msg)
        return data
    data["hit_type"] = 4 * data["hit_top"] + 2 * data["hit_mid"] + data["hit_btm"]
    return data


def resample_data(data: pd.DataFrame, interval: int):
    """測定データをリサンプル

    ``data``に指定した測定データを``pandas.DataFrame.resample``でリサンプルします。
    データは前処理済みのデータを指定してください。
    再集計したデータをさらに、再集計することもできます。

    :Note:
        - 入力データに必要なカラム: ``["time", "hit_type", "hit_top", "hit_mid", "hit_btm", "adc", "tmp", "atm", "hmd"]``
        - 出力データに追加されるカラム: ``["interval", "event_rate", "event_rate_top", "event_rate_mid", "event_rate_btm"]``


    :Args:
        - `data (pd.DataFrame)`: 前処理済みのデータフレーム
        - `interval (int)`: リサンプル間隔（秒）

    :Returns:
        - `merged (pd.DataFrame)`: 再集計後のデータフレーム
    """

    # 再集計する時間間隔
    rule = f"{interval}s"

    # 元データのコピーを作成
    copied = data.copy()
    # インデックスを datetime オブジェクトに変更
    copied.index = copied["time"]

    # count
    headers = {"hit_type": "events"}
    keys = list(headers.keys())
    _count = copied.resample(rule)[keys].count().reset_index().rename(columns=headers)

    # sum: 合計値
    headers = {
        "hit_top": "hit_top_sum",
        "hit_mid": "hit_mid_sum",
        "hit_btm": "hit_btm_sum",
    }
    keys = list(headers.keys())
    _sum = copied.resample(rule)[keys].sum().reset_index()

    # mean: 平均値
    headers = {
        "adc": "adc_mean",
        "tmp": "tmp_mean",
        "atm": "atm_mean",
        "hmd": "hmd_mean",
    }
    keys = list(headers.keys())
    _mean = copied.resample(rule)[keys].mean().reset_index()

    # std : 標準偏差
    headers = {
        "adc": "adc_std",
        "tmp": "tmp_std",
        "atm": "atm_std",
        "hmd": "hmd_std",
    }
    keys = list(headers.keys())
    _std = copied.resample(rule)[keys].std().reset_index().rename(columns=headers)

    # 測定日時（time）で結合
    m1 = pd.merge(_count, _sum, on="time")
    m2 = pd.merge(m1, _mean, on="time")
    merged = pd.merge(m2, _std, on="time")

    # レートを計算
    merged["interval"] = interval
    merged["event_rate"] = merged["events"] / merged["interval"]
    merged["event_rate_top"] = merged["hit_top"] / merged["interval"]
    merged["event_rate_mid"] = merged["hit_mid"] / merged["interval"]
    merged["event_rate_btm"] = merged["hit_btm"] / merged["interval"]

    # 測定時間を計算
    epoch = merged["time"].min()
    merged["days"] = merged["time"] - epoch
    merged["seconds"] = merged["days"].dt.total_seconds()

    return merged


def resample_data_with_hit_type(data: pd.DataFrame, interval: int) -> pd.DataFrame:
    """測定データをリサンプル

    ヒット種類（``hit_type``）ごとにリサンプルします。

    :Args:
        - `data (pd.DataFrame)`: ``hit_type``を持つデータフレーム
        - `interval (int)`:  リサンプルの間隔（秒）

    :Returns:
        - `resampled (pd.DataFrame)`: ``hit_type``ごとにリサンプルしたデータフレーム
    """
    hit_types = sorted(data["hit_type"].unique())
    rdata = []
    for ht in hit_types:
        q = f"hit_type == {ht}"
        qdata = data.query(q).copy()
        _resampled = resample_data(qdata, interval)
        _resampled["hit_type"] = ht
        rdata.append(_resampled)

    resampled = pd.concat(rdata, ignore_index=True)
    resampled = resampled.sort_values(["time", "hit_type"])
    return resampled


def preprocess_data(
    data: pd.DataFrame, interval: int, datetime_offset: int, timezone: str
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """測定データを前処理する

    :Args:
        - `data (pd.DataFrame)`: 測定データ
        - `interval (int)`: リサンプル間隔（秒）
        - `datetime_offset (int)`: 時刻のオフセット（秒）
        - `timezone (str)`: タイムゾーン情報

    :Returns:
        - `data (pd.DataFrame)`: 前処理したあとのデータ
        - `resampled (pd.DataFrame)`: リサンプルしたデータ
    """
    data = add_time(data, offset=datetime_offset, timezone=timezone)
    data = add_hit(data)
    data = add_hit_type(data)
    resampled = resample_data_with_hit_type(data, interval)
    return data, resampled


def raw2csv(
    fnames: list[Path], interval: int, datetime_offset: int, timezone: str
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """測定データを指定してデータフレームに変換する

    :Args:
        - `fnames (list[Path])`: 前処理するファイル名のリスト
        - `interval (int)`: リサンプル間隔（秒）
        - `datetime_offset (int)`: 時刻のオフセット（秒）
        - `timezone (str)`: タイムゾーン情報

    :Returns:
        - `data (pd.DataFrame)`: 前処理したあとのデータ
        - `resampled (pd.DataFrame)`: リサンプルしたデータ

    """
    if len(fnames) == 0:
        msg = "No files found."
        logger.error(msg)
        sys.exit()

    data = read_data(fnames)
    data, resampled = preprocess_data(data, interval, datetime_offset, timezone)
    return data, resampled


def run2csv(run: RunData) -> tuple[pd.DataFrame, pd.DataFrame]:
    """ランを指定してデータフレームに変換する

    :Args:
        - `run (RunData)`: RunDataオブジェクト

    :Returns:
        - `data (pd.DataFrame)`: 前処理したあとのデータ
        - `resampled (pd.DataFrame)`: リサンプルしたデータ

    :Example:

    ```python
    run = RunData(...)
    data, resampled = run2csv(run)
    print(f"gzip={len(data)}")
    print(f"csv ={len(resampled)}")
    ```

    """
    fnames = get_fnames(read_from=run.read_from, search_pattern=run.srcf)
    interval = run.interval
    datetime_offset = run.datetime_offset
    timezone = "UTC+09:00"

    data, resampled = raw2csv(fnames, interval, datetime_offset, timezone)
    data["runid"] = run.run_id
    data["name"] = run.name
    data["description"] = run.description

    resampled["runid"] = run.run_id
    resampled["name"] = run.name
    resampled["description"] = run.description

    return data, resampled


if __name__ == "__main__":
    fname = "../data/raw_data/20230806_run17/osechi_data_000000.csv"
    fnames = [Path(fname)]
    gzip, csv = raw2csv(fnames, interval=600, datetime_offset=0, timezone="UTC+09:00")
    logger.debug(f"raw2gz  = {len(gzip)}")
    logger.debug(f"raw2csv = {len(csv)}")



from dataclasses import dataclass
from icecream import ic
from loguru import logger
from pathlib import Path
from typing import Optional, Any
import pandas as pd
import sys
import tomli  # Python3.9対応のためtomliを使う
from deprecated import deprecated

# import tomllib  # TODO: いずれtomllibに置き換える
from pydantic import BaseModel

ic.configureOutput(prefix=f"{__name__}: ")


@dataclass
class RunData:
    """
    ランごとの設定

    それぞれのランの情報／設定を整理するためのクラスです。
    ランを走らせた時の設定は、Googleスプレッドシートで管理しています
    （更新・共有の手軽さを考慮してスプレッドシートを採用しました）。

    :TODO:
        - スプレッドシートをCSV形式でダウンロードして、RunDataオブジェクトに変換するツールを作りたい
    """

    run_id: int
    """ラン番号"""

    description: str = ""
    """ランの簡単な説明。詳しい説明はGoogleスプレッドシートで説明"""

    read_from: str = ""
    """測定データが保存されているディレクトリ名"""

    srcf: str = "*.csv"
    """測定データの拡張子。デフォルトはCSV形式"""

    interval: int = 600
    """サンプリング間隔（秒）。デフォルトは600秒（10分）"""

    datetime_offset: int = 0
    """測定器の時刻と実時刻の時間差（秒）。デフォルトは0秒"""

    skip: bool = False
    """データ処理スキップのフラグ。デフォルトはFalse"""

    nfiles: int = 0
    """測定データのファイル数。デフォルトは0"""

    raw2gz: str = ""
    """出力ファイル名。リサンプルなし。指定がない場合は保存しない"""

    raw2csv: str = ""
    """出力ファイル名。リサンプルあり。指定がない場合は保存しない"""

    query: str = ""
    """データの抽出条件。有効なデータを指定する"""

    def __post_init__(self) -> None:
        """RunDataクラスの初期化"""
        self.name = f"Run{self.run_id}"
        try:
            self.fnames = self.get_fnames()
        except FileNotFoundError as e:
            logger.warning(e)
            msg = "self.fnames set to []"
            logger.info(msg)
            self.fnames = []
        self.nfiles = len(self.fnames)

    def get_fnames(self) -> list[Path]:
        """該当するランのファイル名の一覧を取得

        `read_from`と`srcf`の値から、該当するファイルの一覧を、
        `pathlib.Path`オブジェクトのリストとして取得します。

        :Exception:
            - `read_from`ディレクトリが存在しない場合は終了

        :Returns:
            - `fnames (list[Path])`: ファイル名のリスト
        """

        read_from = Path(self.read_from).expanduser()
        srcf = self.srcf

        if not read_from.exists():
            msg = f"No directory found : {read_from}"
            raise FileNotFoundError(msg)

        fnames = sorted(read_from.glob(srcf))
        return fnames

    def _load_gzip(self) -> Optional[pd.DataFrame]:
        """gzipで保存したデータを読み込む

        ``raw2gz``で指定したファイル名を`pd.DataFrame`に変換します。

        :Exception:
            - `raw2gz`ファイルが存在しない場合は終了

        :Returns:
            - `data (pd.DataFrame | None)`: データフレーム
        """
        p = Path(self.raw2gz)
        if p.exists():
            data = pd.read_csv(self.raw2gz, parse_dates=["time"])
            return data
        else:
            msg = f"File not found. {p}"
            raise FileNotFoundError(msg)

    def _load_csv(self) -> Optional[pd.DataFrame]:
        """csvで保存したデータを読み込む

        ``raw2csv``で指定したファイルを``pd.DataFrame`に変換します。

        :Exception:
            - ``raw2csv``ファイルが存在しない場合は終了

        :Returns:
            - `data (pd.DataFrame | None)` : データフレーム
        """
        p = Path(self.raw2csv)
        if p.exists():
            data = pd.read_csv(self.raw2csv, parse_dates=["time"])
            return data
        else:
            msg = f"File not found. {p}"
            raise FileNotFoundError(msg)

    def load_data(self, kind: str) -> Optional[pd.DataFrame]:
        """形式を指定してデータを読み込む

        `kind`で測定データの読み取り形式を指定します。
        指定できる形式は`csv`もしくは`gzip`です。

        :Args:
            - `kind (str)`: 保存したデータの形式。["csv", "gzip"]

        :Returns:
            - `data (pd.DataFrame)`: データフレーム
        """
        if kind == "csv":
            return self._load_csv()
        elif kind == "gzip":
            return self._load_gzip()
        else:
            fmt = ["csv", "gzip"]
            msg = f"Wrong file type : {kind}. Choose from {fmt}."
            raise ValueError(msg)


@dataclass
class Config:
    """（削除予定）ラン設定ファイル用のクラス

    :TODO:
        - RunManagerクラスを別途作成した
        - 重複する箇所は置き換える

    """

    fname: str = "config.toml"
    """設定ファイル。デフォルトは`config.toml`"""

    @deprecated(version="v0.17.2", reason="Replace with RunManager.")
    def __post_init__(self) -> None:
        """（削除予定）
        - ``self.config``
        - ``self.rules``
        - ``self.runs``
        """
        self.config = self.load_config()
        self.rules = self.get_rules()
        self.runs = self.get_runs()
        self.labels = self.get_labels()

    @deprecated(version="v0.17.2", reason="Replace with RunManager.")
    def load_config(self) -> dict:
        """（削除予定）設定ファイルを読み込む

        :Returns:
        - config(dict) : 設定

        :Notes:
        - デフォルトの設定ファイル名は `config.toml`
        - 設定ファイルの名前は変更することができる
        - 設定ファイルが見つからない場合は、エラーを表示してスキップ（早期リターン）する
        - 設定ファイルは辞書型で読み込む
        """
        p = Path(self.fname)

        if not p.exists():
            msg = f"No file found : {p}"
            raise FileNotFoundError(msg)

        with p.open("rb") as f:
            config = tomli.load(f)
        return config

    @deprecated(version="v0.17.2", reason="Replace with RunManager.")
    def get_rules(self) -> dict:
        """（削除予定）イベント条件に関する設定を取得する

        :Returns:
        - rules(dict): {イベント名 : イベント条件}

        :Notes:
        - 設定ファイルの ``[rules]`` セクションの内容を取得する
        - ``条件名 = 条件式`` の辞書型（map型）で定義されている

        """
        rules = self.config.get("rules")
        return rules  # type: ignore

    @deprecated(version="v0.17.2", reason="Replace with RunManager.")
    def get_labels(self) -> Optional[dict]:
        """（削除予定）カラム名を取得する

        ``[labels]`` のセクションに、カラム名に対応した日本語を記述する

        """
        labels = self.config.get("labels")
        return labels

    @deprecated(version="v0.17.2", reason="Replace with RunManager.")
    def get_run(self, run_id: int) -> RunData:
        """（削除予定）ラン番号を指定してラン情報を取得する"""
        for run in self.runs:
            if run.run_id == run_id:
                return run
        msg = f"Run #{run_id} is out of range. Add to config."
        logger.error(msg)
        sys.exit()

    @deprecated(version="v0.17.2", reason="Replace with RunManager.")
    def get_runs(self) -> list[RunData]:
        """（削除予定）ランに関係する設定を取得する

        :Returns:
        - runs(list[RunData]): ランごとに設定をまとめたリスト

        :Notes:
        - 設定ファイルの ``[[rawdata]]`` セクションの内容を ``RunData`` クラスにまとめる
        - 除外するランがある場合は ``skip = true`` を設定する
        """
        runs = []
        rundata = self.config["rundata"]
        for data in rundata:
            if data.get("skip") is None:
                run = data.get("run")
                error = f"Run{run} : No configuration found for 'skip'"
                logger.error(error)
                sys.exit()

            if not data.get("skip"):
                _data = RunData(
                    run_id=data["run"],
                    read_from=data["read_from"],
                    srcf=data["srcf"],
                    interval=data["interval"],
                    datetime_offset=data.get("datetime_offset", 0),
                    description=data["desc"],
                    skip=data["skip"],
                    raw2gz=data.get("raw2gz"),
                    raw2csv=data.get("raw2csv"),
                    query=data.get("query"),
                )
                runs.append(_data)
        return runs


@dataclass
class Daq:
    """DAQ設定クラス

    :Example:

    ```python
    daq = Daq()
    daq.load_toml(fname="TOML形式のファイル")

    # 現在日時をベースにした保存先に変更
    now = pendulum.now().format("YYYYMMDD")
    daq.saved =  str(Path(daq.saved) / now)
    ```

    """

    saved: str = "."
    """測定データを保存するディレクトリ。デフォルトは`.`（カレントディレクトリ）"""

    prefix: str = "data"
    """測定データのファイル名につける接頭辞。デフォルトは`data`"""

    suffix: str = ".csv"
    """測定データの拡張子。デフォルトは`.csv`"""

    skip: int = 10

    max_rows: int = 10000
    """1ファイルあたりのデータ数。デフォルトは`10000`"""

    max_files: int = 100
    """1ランあたりのファイル数。デフォルトは`100`"""

    quiet: bool = False
    """quietモード用フラグ。デフォルトはFalse"""

    append: bool = False
    """追記モード用フラグ。デフォルトはFalse"""

    device: str = "/dev/ttyUSB0"
    """シリアル通信のポート名。デフォルトは`/dev/ttyUSB0`（Linux用）"""

    baudrate: int = 115200
    """シリアル通信レート（ボーレート）。デフォルトは`115200` [bps]"""

    timeout: int = 1000
    """シリアル通信のタイムアウト設定（秒）。デフォルトは`1000`秒"""

    fname_logs: str = "threshold_logs.csv"
    """ファイル名。スレッショルド設定のログ"""

    fname_scan: str = "threshold_scan.csv"
    """ファイル名。スレッショルド測定の結果"""

    def load_toml(self, fname: str) -> None:
        "Load DAQ configuration from TOML"
        p = Path(fname)
        with p.open("rb") as f:
            _config = tomli.load(f)

        # logger.debug(_config)
        self.saved = _config.get("saved", ".")
        self.prefix = _config.get("prefix", "data")
        self.suffix = _config.get("suffix", ".csv")
        self.skip = _config.get("skip", 10)
        self.max_rows = _config.get("max_rows", 10000)
        self.max_files = _config.get("max_files", 100)
        self.quiet = _config.get("quiet", False)
        self.append = _config.get("append", False)
        self.device = _config.get("device", "/dev/ttyUSB0")
        self.baudrate = _config.get("baudrate", 115200)
        self.timeout = _config.get("timeout", 1000)
        self.fname_logs = _config.get("fname_logs", "threshold_logs.csv")
        self.fname_scan = _config.get("fname_scan", "threshold_scan.csv")


class UserSettings(BaseModel):
    """ユーザー設定用のクラス

    ``read_from`` に指定したファイルからユーザー設定を読み込みます。
    設定ファイルはTOML形式で作成してください。
    その他の形式に対応する予定はいまのところありません。
    また、読み込み時のファイル形式のチェックもしていません。

    :TODO:
        - ``daq.toml``、``scan.toml``から``hnw.toml``に移行する
        - クラス名を`DaqManager`に変更する

    :versionadded:
        `0.12.0`

    :Example:

    ```python
    us = UserSettings(load_from="../sandbox/hnw.toml")
    us.settings
    {
        'default': {'saved': '', 'suffix': '.csv', 'skip': 10, 'max_rows': 1000},
        'device': {'port': '/dev/ttyUSB0', 'baudrate': 115200, 'timeout': 100},
        'daq': {'prefix': 'osechi_data', 'max_files': 1000},
        'scan': {'prefix': 'scan_data', 'max_files': 10, 'timeout': 10},
        'threshold': {'logs': {'fname': 'threshold_logs.csv',
        'names': ['time', 'ch', 'vth', 'success']},
        'scan': {'fname': 'threshold_scan.csv',
        'names': ['time', 'duration', 'ch', 'vth', 'events']
    }

    us.sections
    dict_keys(
        ['default', 'device', 'daq', 'scan', 'threshold', 'loguru']
    )
    ```

    """

    load_from: str
    """設定ファイルを指定する"""

    settings: dict = {}
    """読み込んだ設定値の一覧"""

    sections: list = []
    """読み込んだ設定セクションの一覧"""

    def model_post_init(self, __context: Any) -> None:
        """UserSettingsクラスの初期化"""
        settings = self.load_toml(self.load_from)
        self.settings = settings
        self.sections = list(settings.keys())
        return super().model_post_init(__context)

    def load_toml(self, load_from: str) -> dict:
        """TOML形式の設定を読み込む

        UserSettingsクラスのオブジェクトを生成するときに ``__post_init__``の中で実行している。
        プロダクション環境では、このメソッドをわざわざ実行する必要はありません。

        新しい設定ファイルを作成した場合の内容確認のために使えると思います。

        :Args:
            - `load_from (str)`: ファイル名

        :Returns:
            - `settings (dict)`: ユーザー設定

        """
        p = Path(load_from)
        with p.open("rb") as f:
            settings = tomli.load(f)
        return settings

    def _get_settings(self, keys: list) -> dict:
        """キーを指定してユーザー設定を取得

        キーはTOMLファイルのセクションに対応しています。
        キーは複数指定できます。
        同じ設定項目がある場合、あとに設定した値が優先されます。

        指定したキーが設定ファイルに存在しない場合はスキップします。

        :Args:
            `keys (list)`: 設定キーをリストで指定する

        :Returns:
            `settings (dict)`: 設定値の一覧

        :Example:

        ```python
        us = UserSettings(load_from="../sandbox/hnw.toml")
        keys = ["default", "device", "scan", "threshold"]
        settings = s._get_settings(keys)
        ```

        """
        settings = {}
        for key in keys:
            d = self.settings.get(key)
            if d is None:
                pass
            else:
                settings.update(d)
        return settings

    def get_daq_settings(self) -> dict:
        """DAQの設定を取得する

        :Note:

            ``keys = ["default", "device", "daq"]``に相当

        :Returns:
            - `settings (dict)`: DAQの設定に必要な項目

        :Example:

        ```python
        us = UserSettings(load_from="../sandbox/hnw.toml")
        settings = us.get_daq_settings()
        settings
        {
            'saved': '',
            'suffix': '.csv',
            'skip': 10,
            'max_rows': 1000,
            'port': '/dev/ttyUSB0',
            'baudrate': 115200,
            'timeout': 100,
            'prefix': 'osechi_data',
            'max_files': 1000
        }
        ```

        """
        keys = ["default", "device", "daq"]
        settings = self._get_settings(keys)
        return settings

    def get_scan_settings(self) -> dict:
        """スレッショルド測定の設定を取得

        :Note:

            ``keys = ["default", "device", "scan", "threshold"]``に相当

        :Returns:
            - `settings (dict)`: スレッショルド測定に必要な項目

        :Example:

        ```python
        us = UserSettings(load_from="../sandbox/hnw.toml")
        settings = us.get_scan_settings()
        settings
        {
            'saved': '',
            'suffix': '.csv',
            'skip': 10,
            'max_rows': 1000,
            'port': '/dev/ttyUSB0',
            'baudrate': 115200,
            'timeout': 10,
            'prefix': 'scan_data',
            'max_files': 10,
            'logs': {'fname': 'threshold_logs.csv', 'names': ['time', 'ch', 'vth', 'success']},
            'scan': {'fname': 'threshold_scan.csv', 'names': ['time', 'duration', 'ch', 'vth', 'events']},
            'fit': {'fname': 'threshold_fit.csv', 'names': ['time', 'ch', '0sigma', '1sigma', '3sigma', '5sigma']}
        }
        ```

        """

        keys = ["default", "device", "scan", "threshold"]
        settings = self._get_settings(keys)
        return settings

    def get_loguru(self, level: str = "DEBUG") -> logger:
        """ロガーを初期化する

        loguruパッケージを使ったロガーを初期化します。
        ``level`` オプションで指定したログレベルで、標準エラー出力（``sys.stderr``）のハンドラーを作成します。

        以下のような ``[loguru]`` セクションを作成することで、ファイル出力のハンドラーを追加できます。
        このハンドラーのログレベルは``DEBUG``、出力形式はJSON形式にハードコードしています。

        ```toml
        [loguru]
        sink = "required"
        format = "required"
        retention = "optional"
        rotation = "optional"
        ```

        :Args:
            - `level (str)`: 標準エラー出力のログレベル。デフォルトは"DEBUG"。

        :Returns:
            - `logger (loguru.logger)`: ロガー

        :Example:

        ```python
        us = UserSettings(load_from="../sandbox/hnw.toml")
        logger = us.get_loguru()
        logger.info("ロガーを初期化した")
        ```
        """
        # ロガーの既定値をリセット
        logger.debug("ロガーをリセットする")
        logger.remove()

        # 標準エラー（sys.stderr）出力のハンドラーを追加
        logger.add(
            sys.stderr,
            format="{time:YYYY-MM-DDTHH:mm:ss} | <level>{level:8}</level> | <level>{message}</level>",
            level=level,
        )
        logger.debug(f"標準エラー出力のハンドラーを追加した（{level=}）")

        # ファイル出力のハンドラーを追加
        section = self.settings.get("loguru")

        if section is None:
            return logger

        # ファイルから読み込んだ値を設定
        logger.add(
            sink=section.get("sink"),
            format=section.get("format"),
            level="DEBUG",
            retention=section.get("retention"),
            rotation=section.get("rotation"),
            serialize=True,
        )
        logger.info("ファイル出力のハンドラーを追加した")
        return logger


@dataclass
class RunManager:
    """ランの設定をスプレッドシートから読み込む"""

    load_from: str
    "ファイル名。データを記録したスプレッドシートを指定する"

    drive: str = "data"
    "ディレクトリ名。上記ファイルまでの相対パスを指定する"

    is_test: bool = False
    "テスト用フラグ。ユニットテストで実行するときはTrueにする"

    def __post_init__(self):
        self.data = self._load_data()
        self.runs = self._get_runs()

    def _load_data(self) -> pd.DataFrame:
        """ファイルから設定を読み込む"""
        data = pd.read_csv(self.load_from, skiprows=2)
        return data

    def get_records(self, query: str, is_valid: bool = True) -> pd.DataFrame:
        """条件にマッチしたレコードを取得

        `query`にマッチしたレコードを取得します。

        :Args:
            - `query (str)`: クエリ条件
            - `is_valid (bool)` : 有効データのフラグ。 Defaults to True

        :Returns:
            - `matched (pd.DataFrame)`: クエリにマッチしたデータフレーム

        :Example:

        ```python
        rm = RunManager("./_data/run.csv")
        rm.get_records("run_id==1")
        rm.get_records("run_type=='古墳'")
        rm.get_records("50 <= run_id <= 84")
        rm.get_records("50 <= run_id <= 84", is_valid=False)
        rm.get_records("50 <= run_id <= 84 and run_type='テスト'")
        ```
        """
        if is_valid:
            matched = self.data.query(query).query("is_valid==True").copy()
        else:
            matched = self.data.query(query).copy()
        return matched

    def _to_rundata(self, row: pd.DataFrame) -> RunData:
        if self.is_test:
            run_data = RunData(run_id=0)
        else:
            run_data = RunData(
                run_id=row.run_id,
                read_from=Path(self.drive) / "raw_data" / row.path_raw_data,
                srcf=row.search_pattern,
                interval=row.resample_interval,
                datetime_offset=row.datetime_offset,
                description=row.overview,
                skip=~row.is_valid,
                raw2gz=Path(self.drive) / "parsed" / row.path_preprocessed_data,
                raw2csv=Path(self.drive) / "parsed" / row.path_resampled_data,
                query=row.query,
            )
        return run_data

    def _get_runs(self, is_valid: bool = True) -> list[RunData]:
        query = "run_id > 0"
        matched = self.get_records(query, is_valid)
        runs = []
        for row in matched.itertuples():
            run_data = self._to_rundata(row)
            runs.append(run_data)
        return runs

    def get_run(self, run_id: int) -> RunData:
        """Get RunData.

        ラン番号（`run_id`）を指定して、RunData情報を取得する。

        :raises KeyError:
            - ラン番号が見つからない場合

        """
        for run in self.runs:
            if run.run_id == run_id:
                return run

        msg1 = f"Run #{run_id} is out of range. Quit."
        logger.error(msg1)
        fname = Path(self.drive) / self.load_from
        msg2 = f"Check or Fix configuration file: {fname}."
        logger.warning(msg2)
        raise KeyError(f"{msg1} {msg2}")


if __name__ == "__main__":
    """configモジュールの動作確認用

    $ python3 config.py

    - 設定ファイルの内容がきちんと読み込まれているか確認する
    - Configクラスのインスタンス変数を修正した場合に動作確認する
    - RunDataクラスのインスタンス変数を修正した場合に動作確認する
    """

    c = Config("../sandbox/config.toml")
    ic(c.fname)
    ic(type(c.rules))
    # ic(c.runs)
    # for run in c.runs:
    #     ic(run.name)
    #     ic(run.runid)
    #     ic(run.fnames)



"""グラフを作成するモジュール（作成途中）"""

import pandas as pd
import altair as alt


def event_rate(data: pd.DataFrame):
    # イベントレート
    hbars = (
        alt.Chart(data)
        .mark_bar(opacity=0.5, color="grey")
        .encode(
            alt.X("time"),
            alt.Y("event_rate").title("イベントレート [Hz]"),
        )
        .properties(width=1200, height=500)
    )

    # 気温
    marks = (
        alt.Chart(data)
        .mark_point(color="blue")
        .encode(
            alt.X("time"),
            alt.Y("tmp").title("気温 [degC]").scale(domain=[20, 35]),
        )
        .properties(width=1200, height=500)
    )

    layers = alt.layer(hbars, marks).resolve_scale(
        y="independent",
    )
    return layers
